"""
Twisted implementation for Kubernetes
"""

from txk8s import _version
from txk8s.lib import (TxKubernetesClient,
                  createPVC,
                  createStorageClass,
                  createDeploymentFromFile,
                  createConfigMap,
                  createService,
                  createServiceAccount,
                  createClusterRole,
                  createClusterRoleBind,
                  createIngress,
                  createEnvVar,)

(_version,
 TxKubernetesClient,
 createPVC,
 createStorageClass,
 createDeploymentFromFile,
 createConfigMap,
 createService,
 createServiceAccount,
 createClusterRole,
 createClusterRoleBind,
 createIngress,
 createEnvVar,) # for pyflakes



"""
Kubernetes Client done Twisted style.
"""
from os import environ

import yaml

from kubernetes import client, config

from twisted.internet import defer, reactor
from twisted.python import log


TIMEOUT = 1 # seconds


class TxKubernetesClient(object):
    """
    Wrapper for Kubernetes Python Client to make requests to the Kubernetes
    API Server asynchronous with Twisted deferreds.
    """
    def __init__(self):
        """
        Load the kube config and create an instance of the Kubernetes API client.
        """
        if environ.get('KUBERNETES_PORT'):
            config.load_incluster_config()
        else:
            config.load_kube_config()

        self.client = client
        self._apiClient = client.ApiClient()
        self.coreV1 = client.CoreV1Api(self._apiClient)
        self.rbacV1Beta1 = client.RbacAuthorizationV1beta1Api(self._apiClient)
        self.extV1Beta1 = client.ExtensionsV1beta1Api(self._apiClient)

    def __getattr__(self, attr):
        """
        Return attributes of the Kubernetes API client.
        """
        return getattr(self.client, attr)

    def call(self, apiMethod, *args, **kwargs):
        """
        Make an asynchronous request to k8s API server with twisted deferreds.
        """
        def _handleErr(fail):
            """
            Log the error and return the failure.
            """
            log.err('Error in: %s. Failure: %s.' % (apiMethod, fail))
            return fail

        d = defer.Deferred()

        # k8s python client v3 supports passing in a callback,
        # but does not handle the error case, therefore addTimeout is added to handle the case of error.
        kwargs['callback'] = d.callback
        apiMethod(*args, **kwargs)
        d.addTimeout(TIMEOUT, reactor)
        d.addErrback(_handleErr)
        return d


def createPVC(res, metadata, spec, namespace):
    """
    Create a Persistent Volume Claim kubernetes resource in a namespace.
    """
    txk8s = TxKubernetesClient()
    body = txk8s.V1PersistentVolumeClaim(
        kind='PersistentVolumeClaim',
        api_version='v1',
        metadata=metadata,
        spec=spec,
    )

    d = txk8s.call(txk8s.coreV1.create_namespaced_persistent_volume_claim,
        namespace,
        body,
    )
    return d
    

def createStorageClass(res, metadata, provisioner):
    """
    Create a Storage Class kubernetes resource.
    """
    txk8s = TxKubernetesClient()
    body = txk8s.V1beta1StorageClass(
        api_version='storage.k8s.io/v1beta1',
        kind='StorageClass',
        metadata=metadata,
        provisioner=provisioner,

    )

    storage = txk8s.StorageV1beta1Api()
    d = txk8s.call(storage.create_storage_class,
        body=body,
    )
    return d


def createDeploymentFromFile(res, filePath, namespace='default'):
    """
    Create a Deployment kubernetes resource from a yaml manifest file.
    """
    txk8s = TxKubernetesClient()

    with open(filePath, 'r') as file:
        deployment = yaml.load(file)

        d = txk8s.call(txk8s.extV1Beta1.create_namespaced_deployment,
            body=deployment,
            namespace=namespace,
        )
        return d


def createConfigMap(metadata, data, namespace):
    """
    Create a configmap kubernetes resources in a namespace.
    """
    txk8s = TxKubernetesClient()

    body = txk8s.V1ConfigMap(
        kind='ConfigMap',
        metadata=metadata,
        data=data,
    )

    d = txk8s.call(txk8s.coreV1.create_namespaced_config_map,
        namespace,
        body
    )
    return d


def createService(filePath, namespace):
    """
    Create a namespaced Service kubernetes resource from a yaml manifest file.
    """
    txk8s = TxKubernetesClient()
    
    with open(filePath, 'r') as file:
        body = yaml.load(file)

        d = txk8s.call(txk8s.coreV1.create_namespaced_service,
            namespace,
            body,
        )
        return d


def createServiceAccount(res, filePath, namespace):
    """
    Create a Service Account kubernetes resource from a yaml manifest file.
    """
    txk8s = TxKubernetesClient()
    
    with open(filePath, 'r') as file:
        body = yaml.load(file)

        d = txk8s.call(txk8s.coreV1.create_namespaced_service_account,
            namespace,
            body,
        )
        return d


def createClusterRole(res, filePath):
    """
    Create a Cluster Role kubernetes resource from a yaml manifest file.
    """
    txk8s = TxKubernetesClient()
    
    with open(filePath, 'r') as file:
        body = yaml.load(file)

        d = txk8s.call(txk8s.rbacV1Beta1.create_cluster_role,
            body,
        )
        return d


def createClusterRoleBind(res, filePath):
    """
    Create a Cluster Role Binding kubernetes resource from a yaml manifest file.
    """
    txk8s = TxKubernetesClient()
    
    with open(filePath, 'r') as file:
        body = yaml.load(file)

        d = txk8s.call(txk8s.rbacV1Beta1.create_cluster_role_binding,
            body,
        )
        return d


def createIngress(res, filePath, namespace):
    """
    Create a Ingress kubernetes resource from a yaml manifest file
    """
    txk8s = TxKubernetesClient()
    
    with open(filePath, 'r') as file:
        body = yaml.load(file)

        d = txk8s.call(txk8s.extV1Beta1.create_namespaced_ingress,
            namespace,
            body,
        )
        return d


def createEnvVar(envName, configMapName, configMapKey):
    """
    Create a environment variable kubernetes resource that references
    a value in a configmap.
    """
    txk8s = TxKubernetesClient()
    return txk8s.V1EnvVar(
        name=envName,
        value_from=txk8s.V1EnvVarSource(
            config_map_key_ref=txk8s.V1ConfigMapKeySelector(
                name=configMapName,
                key=configMapKey,
            ),
        ),
    )



from setuptools import setup
from inspect import cleandoc


_version = {}
execfile('txk8s/_version.py', _version)

setup(
  name = 'txk8s',
  packages = ['txk8s', 'txk8s.test'],
  version = _version['__version__'],
  description = 'A Twisted implementation of Kubernetes',
  author = 'Jessica Grebenschikov',
  author_email = 'jessica@bright.md',
  url = 'https://github.com/Brightmd/txk8s',
  keywords = ['twisted', 'kubernetes'],
  classifiers = [],
  scripts = [],
  install_requires=cleandoc('''
    kubernetes>=3.0.0,<3.1.0
    Twisted>=16.0.0,<18.0.0
    PyYAML>=3.10,<4.0.0
    ''').split()
)



"""
Shared fixtures for txk8s pytests.
"""
import os

from pytest import fixture

from mock import patch

from kubernetes import config

from txk8s import lib


@fixture
def kubeConfig():
    """
    Fixture for kubernetes config patch.
    """
    pInCluster = patch.object(config, 'load_incluster_config')
    pKube = patch.object(config, 'load_kube_config')
    with pInCluster, pKube:
        yield


@fixture
def txclient(kubeConfig):
    """
    Fixture to return an instance of the TxKubernetesClient
    class.
    """
    pEnv = patch.dict(os.environ, {'KUBERNETES_PORT': ''})
    with pEnv:
        return lib.TxKubernetesClient()


@fixture
def txclientInCluster(kubeConfig):
    """
    Fixture to return an instance of the TxKubernetesClient
    class.
    """
    pEnv = patch.dict(os.environ, {'KUBERNETES_PORT': 'tcp://hello'})
    with pEnv:
        return lib.TxKubernetesClient()



"""
Tests for osteoblaster's twisted kubernetes module.
"""
import pytest

from kubernetes import client

from mock import Mock, mock_open, patch

from twisted.python import log

from txk8s import lib


def test_initClient(txclient):
    """
    Do I initialize with the correct attributes?
    """
    for attr in ("client", "_apiClient", "coreV1"):
        assert getattr(txclient, attr)


def test_initClientInCluster(txclientInCluster):
    """
    Do I initialize with the correct attributes?
    """
    for attr in ("client", "_apiClient", "coreV1"):
        assert getattr(txclientInCluster, attr)


def test_getAttr(txclient):
    """
    Do I get attributes from the k8s python api client?
    """
    expected = "<class 'kubernetes.client.models.v1_namespace.V1Namespace'>"
    assert str(txclient.__getattr__("V1Namespace")) == expected


@pytest.inlineCallbacks
def test_clientCallSuccess(kubeConfig):
    """
    Check that the `call` method does the following when successful:
    - adds callback to the kwargs passed to the apiMethod
    - calls the apiMethod it is passed
    - does not call the errback handler when successful
    - returns a deferred
    """
    def fakeReadNamespaceSecret(callback):
        callback('happy')
        return

    pApiMethod = patch.object(client,
        'CoreV1Api',
        return_value=Mock(
            read_namespaced_secret=fakeReadNamespaceSecret,
        ),
        autospec=True,
    )
    pErr = patch.object(log, 'err', autospec=True)

    with pErr as mErr, pApiMethod as mApiMethod:
        txclient = lib.TxKubernetesClient()
        res = yield txclient.call(txclient.coreV1.read_namespaced_secret)
        assert mApiMethod.call_count == 1
        assert 'happy' == res
        assert mErr.call_count == 0


@pytest.inlineCallbacks
def test_clientCallError(kubeConfig):
    """
    Check that the `call` method does the following when unsuccessful:
    - when the timeout is triggered the errback is triggered which logs the message about the failure
    """
    pApiMethod = patch.object(client,
        'CoreV1Api',
        return_value=Mock(
            read_namespaced_secret=Mock(),
        ),
        autospec=True,
    )
    pErr = patch.object(log, 'err', autospec=True)
    pTimeout = patch.object(lib, 'TIMEOUT', 0)

    with pErr as mErr, pApiMethod, pTimeout:
        txclient = lib.TxKubernetesClient()
        d = txclient.call(txclient.coreV1.read_namespaced_secret)
        def _check(fail):
            return
        d.addErrback(_check)
        yield d
        assert mErr.call_count == 1


@pytest.inlineCallbacks
def test_createPVC(kubeConfig):
    """
    Do I create a Persistent Volume Claim kubernetes resource in a namespace?
    """
    meta = 'happy'
    spec = 'days'
    namespace = 'grn-se-com'
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pPVC = patch.object(client, 'V1PersistentVolumeClaim')
    pApiMethod = patch.object(client,'CoreV1Api')
    with pPVC as mPVC, pApiMethod, pCall as mCall:
        yield lib.createPVC('a', meta, spec, namespace)
        mPVC.assert_called_once_with(api_version='v1', kind='PersistentVolumeClaim', metadata=meta, spec=spec)
        mCall.assert_called_once()


@pytest.inlineCallbacks
def test_createStorageClass(kubeConfig):
    """
    Do I create a Storage Class kubernetes resource?
    """
    meta = 'happy'
    provisioner = 'aws-efs'
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pStorage = patch.object(client, 'V1beta1StorageClass')
    pApiMethod = patch.object(client,'StorageV1beta1Api')
    with pStorage as mStorage, pApiMethod, pCall as mCall:
        yield lib.createStorageClass('a', meta, provisioner)
        mStorage.assert_called_once_with(api_version='storage.k8s.io/v1beta1', kind='StorageClass', metadata=meta, provisioner=provisioner)
        mCall.assert_called_once()


@pytest.inlineCallbacks
def test_createDeploymentFromFile(kubeConfig):
    """
    Do I create a Deployment kubernetes resource from a yaml manifest file?
    """
    pOpen = patch("__builtin__.open", mock_open(read_data="data"))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'ExtensionsV1beta1Api',
        return_value=Mock(
            create_namespaced_deployment='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createDeploymentFromFile('a', '/path')
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', body='data', namespace='default')


@pytest.inlineCallbacks
def test_createConfigMap(kubeConfig):
    """
    Do I create a configmap kubernetes resources in a namespace?
    """
    meta = 'happy'
    data = 'days'
    namespace = 'grn-se-com'
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pPVC = patch.object(client, 'V1ConfigMap', return_value='thing')
    pApiMethod = patch.object(client,
        'CoreV1Api',
        return_value=Mock(
            create_namespaced_config_map='a',
        ),
        autospec=True,
    )
    with pApiMethod, pPVC, pCall as mCall:
        yield lib.createConfigMap(meta, data, namespace)
        mCall.assert_called_once_with('a', 'grn-se-com', 'thing')


@pytest.inlineCallbacks
def test_createService(kubeConfig):
    """
    Do I create a namespaced Service kubernetes resource from a yaml manifest file?
    """
    namespace = 'grn-se-com'
    fileData = 'data'
    pOpen = patch("__builtin__.open", mock_open(read_data=fileData))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'CoreV1Api',
        return_value=Mock(
            create_namespaced_service='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createService('/path', namespace)
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', namespace, fileData)


@pytest.inlineCallbacks
def test_createServiceAccount(kubeConfig):
    """
    Do I create a Service Account kubernetes resource from a yaml manifest file?
    """
    namespace = 'grn-se-com'
    fileData = 'data'
    pOpen = patch("__builtin__.open", mock_open(read_data=fileData))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'CoreV1Api',
        return_value=Mock(
            create_namespaced_service_account='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createServiceAccount('a', '/path', namespace)
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', namespace, fileData)


@pytest.inlineCallbacks
def test_createClusterRole(kubeConfig):
    """
    Do I create a Cluster Role kubernetes resource from a yaml manifest file?
    """
    fileData = 'data'
    pOpen = patch("__builtin__.open", mock_open(read_data=fileData))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'RbacAuthorizationV1beta1Api',
        return_value=Mock(
            create_cluster_role='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createClusterRole('a', '/path')
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', fileData)


@pytest.inlineCallbacks
def test_createClusterRoleBind(kubeConfig):
    """
    Do I create a Cluster Role Binding kubernetes resource from a yaml manifest file?
    """
    fileData = 'data'
    pOpen = patch("__builtin__.open", mock_open(read_data=fileData))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'RbacAuthorizationV1beta1Api',
        return_value=Mock(
            create_cluster_role_binding='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createClusterRoleBind('a', '/path')
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', fileData)


@pytest.inlineCallbacks
def test_createIngress(kubeConfig):
    """
    Do I create a Ingress kubernetes resource from a yaml manifest file?
    """
    namespace = 'g-se-com'
    fileData = 'data'
    pOpen = patch("__builtin__.open", mock_open(read_data=fileData))
    pCall = patch.object(lib.TxKubernetesClient, 'call')
    pApiMethod = patch.object(client,
        'ExtensionsV1beta1Api',
        return_value=Mock(
            create_namespaced_ingress='a',
        ),
        autospec=True,
    )
    with pApiMethod as mApiMethod, pCall as mCall, pOpen:
        yield lib.createIngress('a', '/path', namespace)
        mApiMethod.assert_called_once()
        mCall.assert_called_once_with('a', namespace, fileData)


def test_createEnvVar(kubeConfig):
    """
    Do I create a environment variable kubernetes resource that references
    a value in a configmap?
    """
    actual = str(lib.createEnvVar('fun!', 'cmName', 'cmKey'))
    assert "'key': 'cmKey'" in actual
    assert "'name': 'cmName'" in actual
    assert "'name': 'fun!'" in actual
